/**
 * balanced2colourability_wormald.c
 * 
 * Copyright (c) 2016-2017 Ghent University
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/times.h>
#include "util.h"

/*************************Defines and global variables*************************/

//#define POPC(x) (__builtin_popcountll(x)) 
#define BIT(i) (1ULL << (i))


int valid_colouring_found = 0;

int num_marked_vertices = 0;

//The neighbourhood of each vertex (the vertex itself is not included in the neighbourhood)
//unsigned long long int neighbourhood[MAXN];


unsigned int marks2[MAXN];
static int markvalue2 = MAXVAL;
#define RESETMARKS2 {int mki; if ((markvalue2 += 1) > MAXVAL) \
      { markvalue2 = 1; for (mki=0;mki<MAXN;++mki) marks2[mki]=0;}}
#define MARK2(v) marks2[v] = markvalue2
#define UNMARK2(v) marks2[v] = markvalue2 - 1
#define ISMARKED2(v) (marks2[v] == markvalue2)

int number_of_components = 0;
int number_of_vertices_marked = 0;
int number_of_vertices_in_component = 0;
int number_of_edges_in_component = 0;


//unsigned long long int current_component_bitvector = 0;

/******************************************************************************/


unsigned long long int num_bisections_generated = 0;

GRAPH current_graph_removed;
ADJACENCY adj_removed;

ADJACENCY adj_marked;

int pathlengths_marked[MAXN] = {0};
int pathlengths_unmarked[MAXN] = {0};

/*********************************Methods**************************************/


/******************************************************************************/

/**
 * DFS algorithm which recursively marks all vertices which can be reached from
 * current_vertex.
 * Complexity: O(|V|+|E|)
 */
static void mark_dfs(unsigned char current_vertex, int use_marked) {
    MARK2(current_vertex);
    //current_component_bitvector |= BIT(current_vertex);
    //ADDELEMENT1(&current_component_bitvector, current_vertex);
    number_of_vertices_marked++;
    number_of_vertices_in_component++;
    
    //fprintf(stderr, "marking %d\n", current_vertex);
    
    int i;
    for(i = 0; i < adj[current_vertex]; i++)
        if(ISMARKED(current_graph[current_vertex][i]) == use_marked) {
            number_of_edges_in_component++;
            if(!ISMARKED2(current_graph[current_vertex][i]))
                mark_dfs(current_graph[current_vertex][i], use_marked);
        }
        //if(ISMARKED(current_graph[current_vertex][i]) == use_marked
        //        && !ISMARKED2(current_graph[current_vertex][i])) {
        //    mark_dfs(current_graph[current_vertex][i], use_marked);
        //}
    
}

/******************************************************************************/

int graphs_induced_by_monochromatic_components_are_acyclic(int use_marked) {
    
    number_of_vertices_marked = 0;
    number_of_components = 0;
    
    int sum_lengths = 0;
    
    RESETMARKS2;
    int i;
    for(i = 0; i < nv; i++)
        if(!ISMARKED2(i) && ISMARKED(i) == use_marked) {
            number_of_vertices_in_component = 0;
            number_of_edges_in_component = 0;
            
            mark_dfs(i, use_marked);
            number_of_components++;
            
            if((number_of_edges_in_component % 2) != 0) {
                fprintf(stderr, "Error: the ne must be even since every edge was counted twice!\n");
                exit(1);
            }
            number_of_edges_in_component /= 2; //Since every edge was counted twice!
            
            //fprintf(stderr, "number_of_vertices_in_component: %d, number_of_edges_in_component: %d\n",
            //        number_of_vertices_in_component, number_of_edges_in_component);                   
            
            //Is not acyclic
            if(number_of_edges_in_component != number_of_vertices_in_component - 1) {
                return 0;
            }
            
            if(use_marked)
                pathlengths_marked[number_of_edges_in_component]++;
            else
                pathlengths_unmarked[number_of_edges_in_component]++;
            
            sum_lengths += (number_of_edges_in_component + 2);
            
            //if(number_of_vertices_in_component > largest_component)
            //    largest_component = number_of_vertices_in_component;
        }
    
    if(number_of_vertices_marked != nv / 2) {
        fprintf(stderr, "Error: invalid nv marked!\n");
        exit(1);
    }
    
    //fprintf(stderr, "sum_lengths: %d, number_of_components: %d\n", sum_lengths, number_of_components);
    
    //ne / 2
    if(sum_lengths != (3 * nv / 4)) { //Is also ok for nv mod 4 = 2 since taking floor!
        return 0;
    }
    //fprintf(stderr, "ok!\n");
    
    return 1;
}

/******************************************************************************/

int contains_unmarked_vertex_with_more_than_two_unmarked_nbrs() {
    
    int i;
    for(i = 0; i < nv; i++)
        if(!ISMARKED(i) && (adj[i] - adj_marked[i]) > 2)
        //if(!ISMARKED(i) && adj_marked[i] == 0) //We assume that the inputgraphs are cubic! Doesn't make any difference!
            return 1;
    
    return 0;
}

/******************************************************************************/

//Important: it is assumed that newly_marked_vertex is marked!
int contains_marked_vertex_with_more_than_two_marked_nbrs_given_vertex(int newly_marked_vertex) {
    
    //newly_marked_vertex is assumed to be marked!
    if(adj_marked[newly_marked_vertex] > 2)
        return 1;
    
    int i;
    for(i = 0; i < adj[newly_marked_vertex]; i++)
        if(ISMARKED(current_graph[newly_marked_vertex][i])
                && adj_marked[current_graph[newly_marked_vertex][i]] > 2)
            return 1;

    return 0;
}

/******************************************************************************/

void create_graph_without_induced_colour_edges() {
    int i, j;
    for(i = 0; i < nv; i++) {
        adj_removed[i] = 0;
        for(j = 0; j < adj[i]; j++)
            if(ISMARKED(i) != ISMARKED(current_graph[i][j]))
                current_graph_removed[i][adj_removed[i]++] = current_graph[i][j];
    }
}

/******************************************************************************/

/**
 * DFS algorithm which recursively marks all vertices which can be reached from
 * current_vertex.
 * Complexity: O(|V|+|E|)
 */
static void mark_dfs_given_graph(unsigned char current_vertex, GRAPH given_graph, ADJACENCY given_adj) {
    MARK2(current_vertex);
    //current_component_bitvector |= BIT(current_vertex);
    //ADDELEMENT1(&current_component_bitvector, current_vertex);
    number_of_vertices_marked++;
    number_of_vertices_in_component++;
    
    //fprintf(stderr, "marking %d\n", current_vertex);
    
    int i;
    for(i = 0; i < given_adj[current_vertex]; i++) {
        number_of_edges_in_component++;
        if(!ISMARKED2(given_graph[current_vertex][i])) {
            mark_dfs_given_graph(given_graph[current_vertex][i], given_graph, given_adj);
        }
    }
    
}

/******************************************************************************/

int connected_components_in_removed_graph_have_same_nv_and_ne() {

    create_graph_without_induced_colour_edges();
    
/*
    fprintf(stderr, "------\n");

    fprintf(stderr, "org graph:\n");
    printgraph(current_graph, adj);

    fprintf(stderr, "marked vertices:\n");
    int i;
    for(i = 0; i < nv; i++)
        if(ISMARKED(i))
            fprintf(stderr, "%d\n", i);

    fprintf(stderr, "graph removed edges\n");
    printgraph(current_graph_removed, adj_removed);    
*/
    
    number_of_vertices_marked = 0;
    number_of_components = 0;
    
    int num_trees = 0;
    
    RESETMARKS2;
    int i;
    for(i = 0; i < nv; i++)
        if(!ISMARKED2(i)) {
            number_of_vertices_in_component = 0;
            number_of_edges_in_component = 0;
            //fprintf(stderr, "start from vertex %d\n", i);
            
            mark_dfs_given_graph(i, current_graph_removed, adj_removed);
            //mark_dfs(i, use_marked);
            number_of_components++;
            if((number_of_edges_in_component % 2) != 0) {
                fprintf(stderr, "Error: the ne must be even since every edge was counted twice!\n");
                exit(1);
            }
            number_of_edges_in_component /= 2; //Since every edge was counted twice          
            
            //fprintf(stderr, "number_of_vertices_in_component: %d, number_of_edges_in_component: %d\n",
            //        number_of_vertices_in_component, number_of_edges_in_component);                   
            
            //Is not same
            if(number_of_edges_in_component != number_of_vertices_in_component) {
                if((nv % 4) == 2 && number_of_edges_in_component == number_of_vertices_in_component - 1
                        && num_trees == 0) {
                    //OK
                    num_trees++;
                } else                
                    return 0;
            }
            
        }
    
    if(number_of_vertices_marked != nv) {
        fprintf(stderr, "Error: invalid nv marked!\n");
        exit(1);
    }
    
    if((nv % 4) == 2 && num_trees != 1) {
        //If it contained more than 1 tree it was already rejected...
        fprintf(stderr, "Error: must contain exactly 1 tree...\n");
        exit(1);
        return 0;
    }

    return 1;
}

/******************************************************************************/

int graphs_induced_by_monochromatic_components_are_isomorphic() {
    
/*
    fprintf(stderr, "org graph:\n");
    printgraph(current_graph, adj);
   
    fprintf(stderr, "marked vertices:\n");
    int i;
    for(i = 0; i < nv; i++)
        if(ISMARKED(i))
            fprintf(stderr, "%d\n", i);    
*/
    
    
/*
    if(contains_vertex_with_more_than_two_marked_nbrs(1)) {
        fprintf(stderr, "Error: contains vertex with more than 2 marked nbrs (is already tested before!)\n");
        exit(1);
    }    
*/
    
    //We only need to consider paths!
    //if(contains_vertex_with_more_than_two_marked_nbrs(1)
    //        || contains_vertex_with_more_than_two_marked_nbrs(0)) {
    //if(contains_vertex_with_more_than_two_marked_nbrs_old(0)) {
    if(contains_unmarked_vertex_with_more_than_two_unmarked_nbrs()) {
        return 0;
    }    
    
    //Since the induced graphs are linear forests, we just need to compare the num paths and their lengths!
    int i;
    for(i = 0; i < nv; i++) {
        pathlengths_marked[i] = 0;
        pathlengths_unmarked[i] = 0;
    }    
    
    //First test if the induced graphs are acyclic (i.e. forests)
    if(!graphs_induced_by_monochromatic_components_are_acyclic(1)
            || !graphs_induced_by_monochromatic_components_are_acyclic(0)) {
        return 0;
    }
    

    if(memcmp(pathlengths_marked, pathlengths_unmarked, sizeof(int) * nv) != 0) {
        //I.e. not isomorphic linear forests!
        return 0;
    }
    

    //Condition 4 Giuseppe
    if(!connected_components_in_removed_graph_have_same_nv_and_ne()) {
        return 0;
    }
    
/*
    fprintf(stderr, "valid bisection found!\n");
    printgraph(current_graph, adj);
   
    fprintf(stderr, "marked vertices:\n");
    //int i;
    for(i = 0; i < nv; i++)
        if(ISMARKED(i))
            fprintf(stderr, "%d\n", i);        
    
    fprintf(stderr, "-------------------------\n");
*/

    //We already know the linear forests are isomorphic!
    return 1;

}

/******************************************************************************/

int vertex_has_more_than_two_marked_neighbours(int vertex) {
    int num_marked_nbrs = 0;
    int j;
    for(j = 0; j < adj[vertex]; j++)
        if(ISMARKED(current_graph[vertex][j]))
            num_marked_nbrs++;
    
    return num_marked_nbrs > 2;
}

/******************************************************************************/

void colour_in_all_possible_ways_recursion(int next_vertex) {
    
    if(valid_colouring_found)
        return;
    
    if(num_marked_vertices > nv / 2) {
        fprintf(stderr, "Error: num marked vertices is too high!\n");
        exit(1);
    }
    
    if(num_marked_vertices == nv / 2) {
        //Test if it is valid colouring
        num_bisections_generated++;
        
        //Test if this is a valid bisection
        //Found a valid bisection, so we can stop!
        if(graphs_induced_by_monochromatic_components_are_isomorphic())
            valid_colouring_found = 1;
        
        return;
    } 
    
    int is_still_valid_colouring_reverse = 1;
    int i;
    for(i = next_vertex; i < nv && is_still_valid_colouring_reverse; i++) {
        MARK(i);
        num_marked_vertices++;
        int j;
        for(j = 0; j < adj[i]; j++)
            adj_marked[current_graph[i][j]]++;
        
        //Helps a bit
        int num_available_vertices = nv - (i + 1);
        if(num_marked_vertices + num_available_vertices < nv / 2)
            is_still_valid_colouring_reverse = 0;
        
        if(is_still_valid_colouring_reverse) {
            int can_continue = 1;
            
            //if(contains_vertex_with_more_than_two_marked_nbrs(1))
            //Only has to check i or its neighbours!
            if(contains_marked_vertex_with_more_than_two_marked_nbrs_given_vertex(i))
                can_continue = 0;
            
            if(can_continue)
                colour_in_all_possible_ways_recursion(i+1);
            
        }
        
        if(valid_colouring_found)
            return;        
        
        UNMARK(i);
        num_marked_vertices--;       
        for(j = 0; j < adj[i]; j++)
            adj_marked[current_graph[i][j]]--;        

    }
    
}

/******************************************************************************/

int contains_valid_bisection() {

/*
    //Compute bitvectors with the neighbourhood of each vertex
    int i, j;
    for(i = 0; i < nv; i++) {
        neighbourhood[i] = 0;
        for(j = 0; j < adj[i]; j++)
            neighbourhood[i] |= BIT(current_graph[i][j]);
    }    
*/
    
    if((nv % 2) != 0) {
        fprintf(stderr, "Error: graph must have an even number of vertices!\n");
        exit(1);
    }
    
    //Since we are testing the Strong Wormald conjecture!
/*
    if((nv % 4) != 0) {
        fprintf(stderr, "Error: graph must have an even number of edges!\n");
        exit(1);
    }    
*/
    
    int i;
    for(i = 0; i < nv; i++)
        adj_marked[i] = 0;
    
    //Marked vertices are coloured black, the unmarked vertices are coloured white
    
    RESETMARKS;
    num_marked_vertices = 0;
    
    //Colour 0 must be coloured black (to avoid equivalent colourings)
    MARK(0);
    num_marked_vertices++;
    for(i = 0; i < adj[i]; i++)
        adj_marked[current_graph[0][i]]++;
    
    valid_colouring_found = 0;
    
    //num_bisections_generated = 0;
    
    colour_in_all_possible_ways_recursion(1);
    
    //fprintf(stderr, "Number of bisections generated: %d\n", num_bisections_generated);
    
    return valid_colouring_found;

}

/******************************************************************************/

/*
 *
 */
int main(int argc, char** argv) {
    

    if(sizeof (unsigned long long int) != 8) {
        fprintf(stderr, "Error: this version relies on 64 bit unsigned long long ints -- sorry.\n");
        exit(1);
    }    

    int i;
    for(i = 1; i < argc; i++) {
        switch(argv[i][0]) {
            default:
            {
                fprintf(stderr, "Error: invalid option: %s\n", argv[i]);
                exit(1);
            }
        }

    }
    
    fprintf(stderr, "Info: program for testing the strong Wormald conjecture\n");
    fprintf(stderr, "Info: special version which also supports nv mod 4 = 2\n");
    
    //For the timing
    struct tms TMS;

    int codelength = 0;
    unsigned char code[MAXN * (MAXN - 1) / 2 + MAXN];
    unsigned long long int number_of_graphs_read = 0;
    unsigned long long int number_of_graphs_written = 0;
    
    int nuller;
    for(; fread(code, sizeof (unsigned char), 1, stdin);) {
        nv = code[0];
        if(nv > MAXN) {
            fprintf(stderr, "Number of vertices is too big (limit is %d) \n", MAXN);
            exit(0);
        }
        if(code[0] == 0)
            nuller = 1;
        else
            nuller = 0;
        codelength = 1;
        while(nuller < nv - 1) {
            code[codelength] = getc(stdin);
            if(code[codelength] == 0) nuller++;
            codelength++;
        }

        number_of_graphs_read++;
        decode_multicode(code, codelength);
        
        //fprintf(stderr, "Processing graph %llu\n", number_of_graphs_read);
        //printgraph(graph);
        
        if(!contains_valid_bisection()) {
            if(fwrite(code, sizeof (unsigned char), codelength, stdout) < codelength) {
                fprintf(stderr, "Error: couldn't write code!\n");
                exit(1);
            }
            number_of_graphs_written++;
        }
    }
    
    times(&TMS);
    unsigned int savetime = (unsigned int) TMS.tms_utime;
    

    fprintf(stderr, "Read %lld graphs\n", number_of_graphs_read);
    fprintf(stderr, "Debuginfo: total num bisections generated: %llu\n", num_bisections_generated);

    fprintf(stderr, "Written %lld graphs which do NOT have a bisection which fulfills conditions 1, 2, 3 and 4 (i.e. which do not have a strong Wormald colouring)\n", number_of_graphs_written);
    fprintf(stderr, "CPU time filter: %.1f seconds.\n", (double) savetime / (double) CLK_TCK);

    return (EXIT_SUCCESS);
}